-- In this project, you will define some basic functions and
-- predicates on the natural numbers and prove some theorems about
-- them. This project has several parts that are more or less
-- independent. Some are easier and some are tricky. Try to complete
-- as much of it as you can. For full marks, do all of questions 1-6.
-- If you can't complete a proof, you can still get partial credit for
-- finishing parts of the proof, but you get at most 50% for that
-- question.
-- 
-- Remember that you may not collaborate on the projects. If you have
-- questions or are stuck, send email to Peter Selinger or Frank Fu.
--
-- You have 2 weeks to do this project. It is due on February 10.

open import Equality

-- ----------------------------------------------------------------------
-- * Basic definitions

-- We start with the definitions of the natural numbers and addition
-- and multiplication. You have seen these before. For convenience, we
-- use infix notation for addition and multiplication.

infixl 7 _+_
infixl 9 _*_

data ℕ : Set where
  zero : ℕ
  succ : ℕ -> ℕ

{-# BUILTIN NATURAL ℕ #-}

_+_ : ℕ -> ℕ -> ℕ
zero + y = y
succ x + y = succ (x + y)

_*_ : ℕ -> ℕ -> ℕ
zero * m = zero
succ n * m = n * m + m

-- ----------------------------------------------------------------------
-- * More properties of addition and multiplication

-- Note: the proof of the following theorems may require additional
-- lemmas. You have to state and prove your own lemmas. If you need
-- any lemmas from Homework 1 or from class, copy them and their
-- proofs here. (You may have to adjust the notations accordingly).

-- 1. Prove the following distributivity law.
theorem-right-distributive : ∀ (n m k : ℕ) -> (n + m) * k == n * k + m * k
theorem-right-distributive = {!!}

-- 2. State and prove a theorem that asserts the associatitivy of
-- multiplication.
theorem-mult-associative : {!!}
theorem-mult-associative = {!!}

-- ----------------------------------------------------------------------
-- * The "less than or equal" relation

-- We define the "≤" relation as an inductive predicate.  In this way,
-- whenever n,m are natural numbers, n ≤ m will be a set, namely, the
-- set of proofs of n ≤ m.
--
-- You will learn more about inductive predicates on Friday, Jan 29.
-- For now, all you need to know is that the ≤ relation is defined as
-- the smallest relation satisfying the following two properties:
--
-- * Zero is the smallest number:
--   ∀ (n : ℕ) -> zero ≤ n
--
-- * The successor function is monotone:
--   ∀ (n m : ℕ) -> n ≤ m -> succ n ≤ succ m
--
-- The proofs of these properties are called z≤n and s≤s,
-- respectively. Remember that "z≤n" and "s≤s" are identifiers.

infix 5 _≤_

data _≤_ : ℕ -> ℕ -> Set where
  z≤n : ∀ (n : ℕ) -> zero ≤ n
  s≤s : ∀ (n m : ℕ) -> n ≤ m -> succ n ≤ succ m

-- 3. Prove: the relation ≤ is reflexive.
theorem-refl≤ : ∀ (n : ℕ) -> n ≤ n
theorem-refl≤ n = {!!}

-- 4. Prove: the relation ≤ is transitive. Hint: do a case split on
-- the hypothesis n ≤ m.
theorem-trans : ∀ (n m k : ℕ) -> n ≤ m -> m ≤ k -> n ≤ k
theorem-trans = {!!}

-- 5. Prove: the relation ≤ is antisymmetric.
anti-symmetric : ∀ (n m : ℕ) -> n ≤ m -> m ≤ n -> n == m
anti-symmetric = {!!}

-- ----------------------------------------------------------------------
-- * Monotonicity of addition

-- Here, you will prove the monotonicity of addition. You will
-- probably need several lemmas. You must state and prove your own
-- lemmas.
--
-- Hint: you might first want to prove simpler properties, such as
-- n ≤ n' implies n + m ≤ n' + m.

-- 6. Prove: addition is monotone.
theorem-add-monotone : ∀ (n n' m m' : ℕ) -> n ≤ n' -> m ≤ m' -> n + m ≤ n' + m'
theorem-add-monotone = {!!}

-- ----------------------------------------------------------------------
-- * Bonus question: monotonicity of multiplication

-- Here, you will prove the monotonicity of multiplication. This is
-- harder than for addition, and you will need more additional lemmas.

-- 7. Prove: multiplication is monotone.
theorem-mult-monotone : ∀ (n n' m m' : ℕ) -> n ≤ n' -> m ≤ m' -> n * m ≤ n' * m'
theorem-mult-monotone = {!!}
